<?php

/**
 * JCH Optimize - Aggregate and minify external resources for optmized downloads
 *
 * @author    Samuel Marshall <sdmarshall73@gmail.com>
 * @copyright Copyright (c) 2010 Samuel Marshall
 * @license   GNU/GPLv3, See LICENSE file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

// Protect from unauthorized access
defined( '_JEXEC' ) or die();

use JchOptimize\Container;
use JchOptimize\Core\Admin\Tasks;
use JchOptimize\Model\Cache;
use Joomla\CMS\Cache\Cache as JCache;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Installer\Adapter\ComponentAdapter;
use Joomla\CMS\Filesystem\File;
use Joomla\Filesystem\Folder;
use Joomla\Registry\Registry;

class Com_JchoptimizeInstallerScript
{
	protected $removeFiles = [
		'files'   => [],
		'folders' => [
			'administrator/components/com_jchoptimize/cache',
			'administrator/components/com_jchoptimize/Controller',
			'administrator/components/com_jchoptimize/Dispatcher',
			'administrator/components/com_jchoptimize/Helper',
			'administrator/components/com_jchoptimize/Model',
			'administrator/components/com_jchoptimize/Platform',
			'administrator/components/com_jchoptimize/sql',
			'administrator/components/com_jchoptimize/Toolbar',
			'administrator/components/com_jchoptimize/View'
		]

	];

	protected $settingsMap = [
		'pro_remove_unused_js_enable' => 'pro_reduce_unused_js_enable',
		'pro_remove_unused_css'       => 'pro_reduce_unused_css',
		'pro_optimize_gfont_enable'   => 'pro_optimizeFonts_enable'
	];

	/**
	 * Runs after install, update or discover_update
	 *
	 * @param   string            $type  install, update or discover_update
	 * @param   ComponentAdapter  $parent
	 *
	 * @return void
	 */
	public function postflight( string $type, ComponentAdapter $parent ): void
	{
		if ( ! in_array( $type, [ 'install', 'update' ] ) )
		{
			return;
		}

		if ( version_compare( JVERSION, '3.99.99', '<=' ) )
		{
			$config_j3 = $parent->getParent()->getPath( 'source' ) . '/backend/config_j3.xml';
			$config    = JPATH_ADMINISTRATOR . '/components/com_jchoptimize/config.xml';

			File::delete( $config );
			if ( ! File::copy( $config_j3, $config ) )
			{
				$msg = "<p>Couldn't copy the config.xml file</p>";
				JLog::add( $msg, JLog::WARNING, 'jerror' );
			}

			File::delete( JPATH_ADMINISTRATOR . '/components/com_jchoptimize/config_j3.xml' );
			File::delete( JPATH_ADMINISTRATOR . '/components/com_jchoptimize/config_j4.xml' );
		}

		// Remove obsolete files and folders
		$this->removeFilesAndFolders( $this->removeFiles );
		//Update new settings
		$this->updateNewSettings();
	}

	/**
	 * Runs on uninstallation
	 *
	 * @param   ComponentAdapter  $parent  Parent object
	 *
	 * @return  void
	 */
	public function uninstall( ComponentAdapter $parent ): void
	{
		// Clean up Htaccess file
		@include_once( JPATH_ADMINISTRATOR . '/components/com_jchoptimize/autoload.php' );
		Tasks::cleanHtaccess();
		Folder::delete( JPATH_ROOT . '/images/jch-optimize' );
		Folder::delete( JPATH_ROOT . '/jchoptimizecapturecache' );

		$container  = Container::getInstance();
		$cacheModel = $container->get( Cache::class );
		$cacheModel->cleanCache();
	}

	/**
	 * Removes obsolete files and folders
	 *
	 * @param   array  $removeList  The files and directories to remove
	 */
	private function removeFilesAndFolders( array $removeList )
	{
		foreach ( $removeList['files'] ?? [] as $file )
		{
			$f = JPATH_ROOT . '/' . $file;

			@is_file( $f ) && File::delete( $f );
		}

		foreach ( $removeList['folders'] ?? [] as $folder )
		{
			$f = JPATH_ROOT . '/' . $folder;

			@is_dir( $f ) && Folder::delete( $f );
		}
	}

	private function updateNewSettings()
	{
		$params = ComponentHelper::getParams( 'com_jchoptimize' );

		//Some settings have changed
		foreach ( $this->settingsMap as $old => $new )
		{
			if ( ! is_null( $params->get( $old ) ) )
			{
				if ( is_array( $new ) )
				{
					foreach ( $new as $value )
					{
						$params->set( $value, $params->get( $old ) );
					}
				}
				else
				{
					$params->set( $new, $params->get( $old ) );
				}

				$params->remove( $old );
			}
		}

		try
		{
			$this->saveComponentSettings( $params );
			$this->clearSystemCacheGroup();
		}
		catch ( Exception $e )
		{
		}
	}

	/**
	 * @param   Registry  $params
	 *
	 * @return void
	 */
	private function saveComponentSettings( Registry $params )
	{
		$db = Factory::getDbo();
		//Save plugin's params to component
		$query = $db->getQuery( true )
			->update( $db->qn( '#__extensions' ) )
			->set( $db->qn( 'params' ) . ' = ' . $db->q( $params->toString() ) )
			->where( $db->qn( 'element' ) . ' = ' . $db->q( 'com_jchoptimize' ) )
			->where( $db->qn( 'type' ) . ' = ' . $db->q( 'component' ) );
		$db->setQuery( $query );
		$db->execute();
	}

	private function clearSystemCacheGroup()
	{
		$cache = JCache::getInstance( 'callback', [ 'defaultgroup' => '_system' ] );
		$cache->clean();
	}
}